/*
 * *****************************************************************
 * *                                                               *
 * *    Copyright (c) Digital Equipment Corporation, 1991, 1995    *
 * *                                                               *
 * *   All Rights Reserved.  Unpublished rights  reserved  under   *
 * *   the copyright laws of the United States.                    *
 * *                                                               *
 * *   The software contained on this media  is  proprietary  to   *
 * *   and  embodies  the  confidential  technology  of  Digital   *
 * *   Equipment Corporation.  Possession, use,  duplication  or   *
 * *   dissemination of the software and media is authorized only  *
 * *   pursuant to a valid written license from Digital Equipment  *
 * *   Corporation.                                                *
 * *                                                               *
 * *   RESTRICTED RIGHTS LEGEND   Use, duplication, or disclosure  *
 * *   by the U.S. Government is subject to restrictions  as  set  *
 * *   forth in Subparagraph (c)(1)(ii)  of  DFARS  252.227-7013,  *
 * *   or  in  FAR 52.227-19, as applicable.                       *
 * *                                                               *
 * *****************************************************************
 */
/*
 * HISTORY
 */
/*
 * @(#)$RCSfile: mqueue.h,v $ $Revision: 1.1.10.2 $ (DEC) $Date: 1995/03/17 16:34:43 $
 */
/*
 *
 *  Description
 *      This module defines functions, structures, and constants needed to
 *      support POSIX 1003.4 D14 IPC Message Passing.
 *	It is included from user applications that use IPC, and also from
 *	kern_mman.c for the kernel sleep/wakeup functions for IPC.
 *
 */

#ifdef _POSIX_4SOURCE

#ifndef _MQUEUE_H
#define _MQUEUE_H 1

#include <sys/limits.h>
#include <sys/psx4_nspace_ts.h>
#include <sys/signal.h>

#define P4IPC_REGISTER	 1
#define P4IPC_UNREGISTER 2
#define P4IPC_NOTIFY	 3

/*
 * A generic doubly-linked list (queue)
 */
struct psx4queue_entry {
        u_long  next_off;  /* offset to next element's hdr (psx4queue_entry)*/
        u_long  prev_off;  /* offset to prev element's hdr (psx4queue_entry)*/
}psx4queue_entry_t;
 
typedef int mqd_t;   /* Process-level message queue descriptor */

/* MQ_NONBLOCK MUST be same as O_NONBLOCK, since I use either/or 
 * may have to include some other file in order to pick up O_NONBLOCK 
 */
#define MQ_NONBLOCK O_NONBLOCK

#define  P4IPC_MSGSIZE	128  /* Implementation specified default size in bytes of
		       	      * a message.  Used when attr is not specified by 
     			      * the user in mq_open().
			      */
#define  P4IPC_MAXMSG	64   /* Implementation specified number of messages
			      * in a queue.  Used  when attr is not specified 
			      * by the user in mq_open().
			      */

typedef struct mq_attr
    {
    long mq_maxmsg;	/* max # of messages allowed in MQ */
    long mq_msgsize;  	/* max size of any one message */
    long mq_flags;	/* actions and state for mq operations */
    long mq_curmsgs;	/* # of messages currently on mq */
    }mq_attr_t;


/* 
 * Message Queue Header 
 * NOTE: All messages get put on the free queue at queue creation.
 *  A sender grabs a message buffer from the free queue, and a receiver grabs
 *  a message from the used queue.  When a message gets sent, a free buffer
 *  is removed from the free list, filled in, and inserted onto the used
 *  queue.  When the message is received, it is taken from the used queue
 *  and put onto the free queue.
 */
typedef struct p4ipc_qhdr
    {
    rtlock_t      qlock;		/* Lock for queue access 	*/
    long	  num_blocked_senders; 	/* Number of blocked senders 	*/
    long	  num_blocked_receivers;/* Number of blocked senders 	*/
    struct psx4queue_entry mq_free; 	/* Free message queue header, the 
				     	 *  address of which blocked senders 
					 *  wait on */
    struct psx4queue_entry mq_used; 	/* Used message queue header, the 
				     	 *  address of which blocked receivers 
					 *  wait on */
    char	  qname[MQ_NAME_MAX];	/* Pathname of associated object   */
    int		  qsize;	  	/* Total size of queue, incl. msgs */
    mq_attr_t 	  qattr;	  	/* Message queue attribute struct  */
    pid_t	  qsignal_pid;		/* process id of process that has
				 	 *  a notification request attached 
				 	 *  to the queue. */
    struct   sigevent	qsigevent; /* Structure containing signal information*/
    void	*qsiginfo;	/* Address of Realtime signal information
				 *  structure to notify a process when
				 *  the queue becomes non-empty, IF the
				 *  siginfo address is stored in the queue
				 *  when realtime signals are implemented.
				 *  This is left here in case it is needed
				 *  when realtime signals are supported.
				 */
    long	last_msg_offset;/* For range checking on prev/next ptrs
				  *  within message headers, this is the
				  *  offset to the last valid message hdr,
				  *  anything beyond this is a bogus ptr.
				  *  Remember that the queue is preallocated,
				  *  so we know what address range things
				  *  will reside in. CURRENTLY NOT USED.
				  */
    int	     	qhdr_reserved[4];  /* Reserved */
    } p4ipc_qhdr_t;

/* 
 * Message header
 *  Each message data buffer is be permanently attached to one 
 *  message header. 
 */
typedef struct p4ipc_mhdr
    {
    u_long  next_off;		/* offset to next message's hdr (psx4queue_entry) */
    u_long  prev_off;       	/* offset to prev message's hdr (psx4queue_entry) */
    unsigned int 	mprio;  /* Message priority */
    long		msize;  /* Size of message buffer */
    int	     	mhdr_reserved[4];  /* Reserved */
    } 
p4ipc_mhdr_t;


#ifndef _KERNEL
typedef struct psx4queue_entry    *mq_queue_t;
typedef struct psx4queue_entry    *mq_queue_entry_t;

/*
 *      Macro:          mq_init
 *      Function:
 *              Initialize the given queue.
 *      Header:
 *              void mq_init(q)
 *                      struct psx4queue_entry *q;
 */
#define mq_init(q)   ((q)->next_off = (q)->prev_off = 0)


/*
 *      Macro:          mq_first
 *      Function:
 *              Returns the address of the first entry in the queue
 *      Header:
 *              mq_queue_entry_t mq_first(q)
 *                      mq_queue_t q;
 */
#define mq_first(q)  (((u_long)q) + ((q)->next_off))


/*
 *      Macro:          mq_last
 *      Function:
 *              Returns the address of the last entry in the queue
 *      Header:
 *              mq_queue_entry_t mq_last(q)
 *                      mq_queue_t q;
 */
#define mq_last(q)  (((u_long)q) + ((q)->prev_off))


/*
 *      Macro:          mq_next
 *      Function:
 *              Returns the address of the next entry after the
 *              specified entry
 *      Header:
 *              mq_queue_entry_t mq_next(q)
 *                      mq_queue_t q_head;
 *                      mq_queue_t q;
 */
#define mq_next(q_head,q)  (((u_long)q_head) + ((q)->next_off))


/*
 *      Macro:          mq_end
 *      Function:
 *              Returns TRUE if the entries are equal
 *      Header:
 *              boolean_t mq_end(q, qe)
 *                      mq_queue_t q;
 *                      mq_queue_entry_t qe;
 */
#define mq_end(q, qe)   ((q) == (qe))


/*
 *      Macro:          mq_empty
 *      Function:
 *              Returns TRUE if the queue is empty.
 *      Header:
 *              boolean_t mq_empty(q)
 *                      mq_queue_t q;
 */
#define mq_empty(q)     mq_end( ((u_long)q), ((u_long)mq_first(q)) )
#endif /* _KERNEL */


/* POSIX 1003.4 Message passing functions */
#ifdef _NO_PROTO

mqd_t mq_open();
int mq_close();
int mq_unlink();
int mq_notify();
int mq_send();
ssize_t mq_receive();
int mq_getattr();
int mq_setattr();

#else
/*
 * Routine definitions
 */
mqd_t mq_open(const char *name, int oflag, ...);
int mq_close(mqd_t mqdes);
int mq_unlink(const char *name);
int mq_notify(mqd_t mqdes, const struct sigevent *notification);
int mq_send(mqd_t mqdes,const char *msg_ptr,size_t msg_len,unsigned int msg_prio);
ssize_t mq_receive(mqd_t mqdes,char *msg_ptr,size_t msg_len,unsigned int *msg_prio);
int mq_getattr(mqd_t mqdes,const struct mq_attr *mqstat);
int mq_setattr(mqd_t mqdes,const struct mq_attr *mqstat,struct mq_attr *omqstat);

#endif  /* _NO_PROTO_ */
#endif  /* _MQUEUE_H */
#endif  /* _POSIX_4SOURCE */

